package org.jastadd.ragconnect.compiler;

import beaver.Parser;
import org.jastadd.option.BooleanOption;
import org.jastadd.option.ValueOption;
import org.jastadd.ragconnect.ast.*;
import org.jastadd.ragconnect.parser.RagConnectParser;
import org.jastadd.ragconnect.scanner.RagConnectScanner;
import org.jastadd.relast.compiler.AbstractCompiler;
import org.jastadd.relast.compiler.CompilerException;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Compiler extends AbstractCompiler {

//  private ValueOption optionOutputDir;
  private ValueOption optionRootNode;
  private ValueOption optionProtocols;
  private BooleanOption optionPrintYaml;
  private BooleanOption optionVerbose;
  private BooleanOption optionLogReads;
  private BooleanOption optionLogWrites;
  private BooleanOption optionLogIncremental;
  private BooleanOption optionExperimentalJastAdd329;

  private static final String OPTION_PROTOCOL_MQTT = "mqtt";
  private static final String OPTION_PROTOCOL_REST = "rest";
  private static final String OPTION_PROTOCOL_JAVA = "java";

  private final static Logger LOGGER = Logger.getLogger(Compiler.class.getName());

  public Compiler() {
    super("ragconnect", true);
  }

  @Override
  protected int compile() throws CompilerException {
    if (getConfiguration().shouldPrintVersion()) {
      System.out.println(readVersion());
      return 0;
    }
    if (getConfiguration().shouldPrintHelp()) {
      getConfiguration().printHelp(System.out);
      return 0;
    }
    if (optionVerbose.value()) {
      LOGGER.setLevel(Level.FINE);
    }

    LOGGER.info(() -> "Running RagConnect " + readVersion());

    if (!getConfiguration().outputDir().exists()) {
      try {
        Files.createDirectories(getConfiguration().outputDir().toPath());
      } catch (IOException e) {
        throw new CompilerException("Error creating output dir " + getConfiguration().outputDir(), e);
      }
    }

    if (!optionRootNode.isMatched()) {
      return error("Root node not specified");
    }

    RagConnect ragConnect = parseProgram(getConfiguration().getFiles());

    // lets inspect ragconnects tree
    /*
    for(EndpointDefinition ed : ragConnect.getEndpointDefinitionList()){
      if(ed.isReceiveTokenEndpointDefinition()){
        for(MappingDefinition md : ed.asReceiveTokenEndpointDefinition().effectiveMappings()){
          //     if (!effectiveMappings().get(effectiveMappings().size() - 1).getToType().assignableTo(
          //              getToken().effectiveJavaTypeUse())) {
          System.out.println("gtt: " + md.getToType());
        }
      }
    }
    */
    if (!ragConnect.errors().isEmpty()) {
      StringBuilder sb = new StringBuilder("Errors:\n");
      for (ErrorMessage e : ragConnect.errors()) {
        sb.append(e).append("\n");
      }
      LOGGER.severe(sb::toString);
      System.exit(1);
    }

    if (optionPrintYaml.value()) {
      ASTNode.rootNode = ragConnect.getProgram().resolveTypeDecl(optionRootNode.value());
      String yamlContent = ragConnect.toMustache().toYAML().prettyPrint();
      System.out.println(yamlContent);
      writeToFile(getConfiguration().outputDir().toPath().resolve("RagConnect.yml"), yamlContent);
      return 0;
    }

    LOGGER.fine("Writing output files");
    final List<String> handlers = new ArrayList<>();
    if (ASTNode.usesMqtt) {
      handlers.add("MqttHandler.jadd");
    }
    if (ASTNode.usesJava) {
      handlers.add("JavaHandler.jadd");
    }
    if (ASTNode.usesRest) {
      handlers.add("RestHandler.jadd");
    }
    // copy handlers into outputDir
    for (String handlerFileName : handlers) {
      try {
        InputStream inputStream = Compiler.class.getClassLoader().getResourceAsStream(handlerFileName);
        if (inputStream == null) {
          throw new CompilerException("Could not open " + handlerFileName);
        }
        Files.copy(inputStream,
            getConfiguration().outputDir().toPath().resolve(handlerFileName),
            StandardCopyOption.REPLACE_EXISTING);
      } catch (IOException e) {
        throw new CompilerException("Could not copy " + handlerFileName, e);
      }
    }
    for (GrammarFile grammarFile : ragConnect.getProgram().getGrammarFileList()) {
      Path outputFile = getConfiguration().outputDir().toPath().resolve(grammarFile.getFileName());
      writeToFile(outputFile, grammarFile.generateAbstractGrammar());
    }
    writeToFile(getConfiguration().outputDir().toPath().resolve("RagConnect.jadd"), ragConnect.generateAspect(optionRootNode.value()));
    return 0;
  }

  public static void main(String[] args) {
    System.setProperty("mustache.debug", "true");
    try {
      new Compiler().run(args);
    } catch (CompilerException e) {
      LOGGER.log(Level.SEVERE, e.getMessage(), e);
      System.exit(1);
    }
  }

  /**
   * Reads the version string.
   *
   * The version string is read from the property file
   * src/main/resources/Version.properties. This
   * file should be generated during the build process. If it is missing
   * then there is some problem in the build script.
   *
   * @author Jesper Öqvist <jesper.oqvist@cs.lth.se>
   * @return the read version string, or <code>version ?</code>
   */
  private String readVersion() {
    try {
      ResourceBundle resources = ResourceBundle.getBundle("ragConnectVersion");
      return resources.getString("version");
    } catch (MissingResourceException e) {
      return "version ?";
    }
  }

  private void writeToFile(Path path, String str) throws CompilerException {
    try (BufferedWriter writer = Files.newBufferedWriter(path)) {
      writer.append(str);
    } catch (Exception e) {
      throw new CompilerException("Could not write to file " + path.toAbsolutePath(), e);
    }
  }

  protected void initOptions() {
    super.initOptions();
    optionRootNode = addOption(
        new ValueOption("rootNode", "root node in the base grammar.")
        .acceptAnyValue()
        .needsValue(true));
    optionProtocols = addOption(
        new ValueOption("protocols", "Protocols to enable")
        .acceptMultipleValues(true)
        .addAcceptedValue(OPTION_PROTOCOL_JAVA, "Enable Java")
        .addAcceptedValue(OPTION_PROTOCOL_REST, "Enable REST")
        .addDefaultValue(OPTION_PROTOCOL_MQTT, "Enable MQTT")
    );
    optionPrintYaml = addOption(
        new BooleanOption("printYaml", "Print out YAML instead of generating files")
        .defaultValue(false));
    optionVerbose = addOption(
        new BooleanOption("verbose", "Print more messages while compiling.")
        .defaultValue(false));
    optionLogReads = addOption(
        new BooleanOption("logReads", "Enable logging for every read.")
        .defaultValue(false));
    optionLogWrites = addOption(
        new BooleanOption("logWrites", "Enable logging for every write.")
        .defaultValue(false));
    optionLogIncremental = addOption(
        new BooleanOption("logIncremental", "Enable logging for observer in incremental dependency tracking.")
            .defaultValue(false));
    optionExperimentalJastAdd329 = addOption(
        new BooleanOption("experimental-jastadd-329", "Use trace events INC_FLUSH_START and INC_FLUSH_END (JastAdd issue #329).")
            .defaultValue(false));
  }

  private RagConnect parseProgram(Collection<String> files) throws CompilerException {
    Program program = new Program();
    boolean atLeastOneGrammar = false;
    boolean atLeastOneRagConnect = false;

    RagConnect ragConnect = new RagConnect();
    ragConnect.setProgram(program);

    GrammarFile ragConnectGrammarPart = new GrammarFile();
    ragConnectGrammarPart.setFileName("RagConnect.relast");
    program.addGrammarFile(ragConnectGrammarPart);

    for (String filename : files) {
      String extension = filename.substring(filename.lastIndexOf('.') + 1);
      switch (extension) {
        case "ast":
        case "relast":
          // processGrammar
          program.addGrammarFile(parseGrammar(filename));
          atLeastOneGrammar = true;
          break;
        case "connect":
        case "ragconnect":
          // process ragConnect
          ragConnect.addConnectSpecificationFile(parseConnectSpec(filename));
          atLeastOneRagConnect = true;
          break;
        default:
          throw new CompilerException("Unknown file extension " + extension + " in " + filename);
      }
    }
    if (!atLeastOneGrammar) {
      LOGGER.severe(() -> "No grammar file specified! (*.ast, *.relast)");
    }
    if (!atLeastOneRagConnect) {
      LOGGER.severe(() -> "No ragconnect file specified! (*.connect, *.ragconnect)");
    }
    if (!atLeastOneGrammar && !atLeastOneRagConnect) {
      System.exit(1);
    }

    // here, the program subtree is also flushed and resolved
    ragConnect.flushTreeCache();
    ragConnect.treeResolveAll();

    ragConnect.additionalRelations().forEach(ragConnectGrammarPart::addDeclaration);
    ragConnect.additionalTokens().forEach(TypeDecl::addComponent);
    ASTNode.loggingEnabledForReads = optionLogReads.value();
    ASTNode.loggingEnabledForWrites = optionLogWrites.value();
    ASTNode.loggingEnabledForIncremental = optionLogIncremental.value();
    ASTNode.experimentalJastAdd329 = optionExperimentalJastAdd329.value();

    // reuse "--incremental" option of JastAdd
    ASTNode.incrementalOptionActive = getConfiguration().incremental() && getConfiguration().traceFlush();
    LOGGER.fine(() -> "ASTNode.incrementalOptionActive = " + ASTNode.incrementalOptionActive);

    // reuse "--List" option of JastAdd
    ASTNode.JastAddList = getConfiguration().listType();

    ASTNode.usesMqtt = optionProtocols.hasValue(OPTION_PROTOCOL_MQTT);
    ASTNode.usesJava = optionProtocols.hasValue(OPTION_PROTOCOL_JAVA);
    ASTNode.usesRest = optionProtocols.hasValue(OPTION_PROTOCOL_REST);
    return ragConnect;
  }

  private GrammarFile parseGrammar(String filename) throws CompilerException {
    try (BufferedReader reader = Files.newBufferedReader(Paths.get(filename))) {
      RagConnectScanner scanner = new RagConnectScanner(reader);
      RagConnectParser parser = new RagConnectParser();
      GrammarFile grammarFile = (GrammarFile) parser.parse(scanner);
      if (optionVerbose.value()) {
        LOGGER.fine(grammarFile::dumpTree);
      }
      grammarFile.setFileName(toBaseName(filename));
      return grammarFile;
    } catch (IOException | Parser.Exception e) {
      throw new CompilerException("Could not parse grammar file " + filename, e);
    }
  }

  private ConnectSpecificationFile parseConnectSpec(String filename) throws CompilerException {
    try (BufferedReader reader = Files.newBufferedReader(Paths.get(filename))) {
      RagConnectScanner scanner = new RagConnectScanner(reader);
      RagConnectParser parser = new RagConnectParser();
      ConnectSpecificationFile specificationFile = (ConnectSpecificationFile) parser.parse(scanner, RagConnectParser.AltGoals.connect_specification_file);
      specificationFile.setFileName(toBaseName(filename));
      return specificationFile;
    } catch (IOException | Parser.Exception e) {
      throw new CompilerException("Could not parse connect file " + filename, e);
    }
  }

  /**
   * Extracts the basename of the given file, with file extension
   * @param filename the given filename
   * @return the basename
   */
  private String toBaseName(String filename) {
    return new File(filename).getName();
  }

//  protected void printUsage() {
//    System.out.println("Usage: java -jar ragconnect.jar [--option1] [--option2=value] ...  <filename1> <filename2> ... ");
//    System.out.println("Options:");
//    System.out.print(commandLine.printOptionHelp());
//  }
}
