package org.jastadd.ragconnect.tests.list;

import org.jastadd.ragconnect.tests.AbstractMqttTest;
import org.jastadd.ragconnect.tests.TestUtils;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.nio.file.Paths;
import java.util.List;
import java.util.function.Function;

import static java.util.Collections.addAll;
import static org.assertj.core.util.Lists.newArrayList;
import static org.jastadd.ragconnect.tests.list.AbstractListTest.IntList.list;
import static org.jastadd.ragconnect.tests.TestUtils.testJaddContainReferenceToJackson;
import static org.junit.jupiter.api.Assertions.assertEquals;

/**
 * Base class for test cases "list manual" and "list incremental".
 *
 * @author rschoene - Initial contribution
 */
@Tag("List")
public abstract class AbstractListTest extends AbstractMqttTest {

  public interface TestWrapperJastAddList<T> extends Iterable<T> {
    int getNumChild();
  }
  public interface TestWrapperReceiverRoot {
    TestWrapperJastAddList<? extends TestWrapperA> getAList();
    TestWrapperJastAddList<? extends TestWrapperA> getAs();
    int getNumA();
    TestWrapperA getA(int index);

    TestWrapperJastAddList<? extends TestWrapperA> getFromSingleAList();
    TestWrapperJastAddList<? extends TestWrapperA> getFromSingleAs();

    TestWrapperJastAddList<? extends TestWrapperA> getWithAddFromAList();
    TestWrapperJastAddList<? extends TestWrapperA> getWithAddFromAs();

    TestWrapperJastAddList<? extends TestWrapperA> getWithAddFromSingleAList();
    TestWrapperJastAddList<? extends TestWrapperA> getWithAddFromSingleAs();

    boolean connectAList(String mqttUri) throws IOException;
    boolean connectFromSingleAList(String mqttUri) throws IOException;
    boolean connectWithAddFromAList(String mqttUri) throws IOException;
    boolean connectWithAddFromSingleAList(String mqttUri) throws IOException;
  }
  public interface TestWrapperA {
    AbstractListTest.TestWrapperB getB(int i);
    int getNumB();
    int getID();
  }
  public interface TestWrapperB {
    int getID();
  }

  AbstractListTest(String shortName) {
    this.shortName = shortName;
  }

  protected static final String TOPIC_A = "a";
  protected static final String TOPIC_SINGLE_A = "single-a";

  protected TestWrapperReceiverRoot receiverRoot;
  protected ReceiverData data;
  protected ReceiverData dataSingle;

  private final String shortName;

  @Test
  public void checkJacksonReference() {
    testJaddContainReferenceToJackson(
        Paths.get("src", "test",
            "02-after-ragconnect", shortName, "RagConnect.jadd"), true);
  }

  @Override
  protected void communicateSendInitialValue() throws InterruptedException {
    checkTree(1, list(), list(0), list(), list(0));

    setInput(1);
    checkTree(2, list(1), list(1), list(1), list(0, 1));

    setInput(1);
    checkTree(2, list(1), list(1), list(1), list(0, 1));

    setInput(2);
    checkTree(3, list(1, 2), list(2), list(1, 1, 2), list(0, 1, 2));

    setInput(3);
    checkTree(4, list(1, 2, 3), list(3), list(1, 1, 2, 1, 2, 3), list(0, 1, 2, 3));
  }

  @Override
  protected void communicateOnlyUpdatedValue() throws InterruptedException {
    checkTree(0, list(), list(), list(), list());

    setInput(1);
    checkTree(1, list(1), list(1), list(1), list(1));

    setInput(1);
    checkTree(1, list(1), list(1), list(1), list(1));

    setInput(2);
    checkTree(2, list(1, 2), list(2), list(1, 1, 2), list(1, 2));

    setInput(3);
    checkTree(3, list(1, 2, 3), list(3), list(1, 1, 2, 1, 2, 3), list(1, 2, 3));
  }

  protected abstract void setInput(int input);

  private void checkTree(int expectedTransmissions, IntList normalA, IntList fromSingleA, IntList withAddA, IntList withAddFromSingleA) throws InterruptedException {
    TestUtils.waitForMqtt();

    assertEquals(expectedTransmissions, data.numberOfElements, "transmissions for normal");
    assertEquals(expectedTransmissions, dataSingle.numberOfElements, "transmissions for single");

    checkList(normalA.toList(), receiverRoot.getNumA(), receiverRoot::getA, true);
    checkList(normalA.toList(), receiverRoot.getAList(), true);
    checkList(normalA.toList(), receiverRoot.getAs(), true);

    checkList(fromSingleA.toList(), receiverRoot.getFromSingleAList(), false);
    checkList(fromSingleA.toList(), receiverRoot.getFromSingleAs(), false);

    checkList(withAddA.toList(), receiverRoot.getWithAddFromAList(), true);
    checkList(withAddA.toList(), receiverRoot.getWithAddFromAs(), true);

    checkList(withAddFromSingleA.toList(), receiverRoot.getWithAddFromSingleAList(), false);
    checkList(withAddFromSingleA.toList(), receiverRoot.getWithAddFromSingleAs(), false);
  }

  private void checkList(List<Integer> expectedList, int numChildren, Function<Integer, TestWrapperA> getA, boolean expectB) {
    assertEquals(expectedList.size(), numChildren, "same list size");
    for (int index = 0; index < expectedList.size(); index++) {
      TestWrapperA a = getA.apply(index);
      assertEquals(expectedList.get(index), a.getID(), "correct ID for A");
      if (expectB) {
        assertEquals(1, a.getNumB(), "one B child");
        assertEquals(expectedList.get(index) + 1, a.getB(0).getID(), "correct ID for B child");
      }
    }
  }

  private void checkList(List<Integer> expectedList, TestWrapperJastAddList<? extends TestWrapperA> actualList, boolean expectB) {
    assertEquals(expectedList.size(), actualList.getNumChild(), "same list size");
    int index = 0;
    for (TestWrapperA a : actualList) {
      assertEquals(expectedList.get(index), a.getID(), "correct ID for A");
      if (expectB) {
        assertEquals(1, a.getNumB(), "one B child");
        assertEquals(expectedList.get(index) + 1, a.getB(0).getID(), "correct ID for B child");
      }
      index++;
    }
  }

  protected static class ReceiverData {
    int numberOfElements = 0;
  }

  protected static class IntList {
    private final List<Integer> integers = newArrayList();
    public IntList(Integer... values) {
      addAll(integers, values);
    }

    public List<Integer> toList() {
      return integers;
    }

    public static IntList list(Integer... values) {
      return new IntList(values);
    }
  }

}
